/*********************************************************************/
/* Mtx : Matrix Package                                              */
/* Description : Main Implementation File                            */
/* ----------------------------------------------------------------- */
/* Author : Benigne DEPREY & Laurent SAINT-MARCEL                    */
/* Date   : 01-Mar-2000                                              */
/*********************************************************************/

/*-------------------------------------------------------------------*/
/* Nested Includes                                                   */
/*-------------------------------------------------------------------*/
#include "matrix.h"
#include "stdio.h"
#include <assert.h>

/*-------------------------------------------------------------------*/
/* Private Procedure Declaration                                     */
/*-------------------------------------------------------------------*/

/*********************************************************************/
/* Mtx_CleanStorage                                                  */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Set all the elements to 0 and initialize the size.              */
/*********************************************************************/
inline void Mtx_CleanStorage (Mtx_Matrix* mtx, int nbLine, int nbColumn)
{
  int i,j;

  mtx->nbLine   = nbLine;
  mtx->nbColumn = nbColumn;

  for(i=0; i<cMATRIX_SIZE; i++) {
    for(j=0; j<cMATRIX_SIZE; j++) {
      (*mtx).tab[i][j] = 0.0;
    }
  }
}

/*-------------------------------------------------------------------*/
/* Implementation                                                    */
/*-------------------------------------------------------------------*/

/*********************************************************************/
/* Mtx_ManagerInit                                                   */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Initialize the manager.                                         */
/*********************************************************************/
void Mtx_ManagerInit ()
{
  int i;
  int size;

  theManager.managerSize = cMANAGER_SIZE;
  theManager.storageLeft = cMANAGER_SIZE;
  size = theManager.managerSize;

  for(i=0; i<(size-1); i++) {
    /* Link with the next cell */
    theManager.storageBuffer[i].next = &theManager.storageBuffer[i+1];

    /* Initialize the matrix */
    Mtx_CleanStorage(&theManager.storageBuffer[i],0,0);
  }
  
  /* Link the last cell with NULL */
  theManager.storageBuffer[size-1].next = NULL;
  
  /* Initialize the last matrix */
  Mtx_CleanStorage(&theManager.storageBuffer[size-1],0,0);

  theManager.freeStorage = theManager.storageBuffer;
}

/*********************************************************************/
/* Mtx_ManagerKill                                                   */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Not used yet.                                                   */
/*********************************************************************/
void Mtx_ManagerKill ()
{

}

#if 1
/*********************************************************************/
/* Mtx_MatrixPrint                                                   */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Print the given matrix.                                         */
/*********************************************************************/
void Mtx_MatrixPrint (Mtx_Matrix* mtx)
{
#ifdef DEBUG
  int i,j;

  for(i=0; i<mtx->nbLine; i++) {
    printf("[ ");    
    for(j=0; j<mtx->nbColumn; j++) {
      printf("%f ",(*mtx).tab[i][j]);
    }
    printf("]\n");
  }
  printf("\n");
#endif
}
#endif

/*********************************************************************/
/* Mtx_CreateIdent                                                   */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Set the diagonal of the given matrix with ones.                 */
/*********************************************************************/
void Mtx_CreateIdent (Mtx_Matrix* mtx)
{
  int i;
  int max = (mtx->nbLine > mtx->nbColumn)?mtx->nbLine:mtx->nbColumn;
  
  for(i=0; i<max; i++) {
    (*mtx).tab[i][i] = 1.0; 
  }
}

/*********************************************************************/
/* Mtx_DeleteMatrix                                                  */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return the given matrix to the manager.                         */
/*********************************************************************/
void Mtx_DeleteMatrix (Mtx_Matrix* mtx)
{
  if (mtx) {

    if(mtx->next!=0) printf("** Manager : espace memoire deja libre !!\n\n\n\n");

    /* Link the freed matrix with the others */
    mtx->next              = theManager.freeStorage;

    /* Anchor the freed matrix */
    theManager.freeStorage = mtx;

    /* Update the number of storage left */
    theManager.storageLeft++;

    //printf("** Manager : [%d cells left]\n",theManager.storageLeft);
  }else{
    printf("** Manager : espace memoire non allou !!!\n");

  }
}

/*********************************************************************/
/* Mtx_GetNewMatrix                                                  */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix given by the manager.                       */
/*********************************************************************/
Mtx_Matrix* Mtx_GetNewMatrix (int nbLine, int nbColumn)
{
  if (theManager.freeStorage) {
    /* Unlink the matrix */
    Mtx_Matrix* newMatrix  = theManager.freeStorage;
    theManager.freeStorage = theManager.freeStorage->next;

    /* Clean the storage */
    newMatrix->next = NULL;
    Mtx_CleanStorage(newMatrix,nbLine,nbColumn);

    /* Update the number of storage left */
    theManager.storageLeft--;

    //printf("** Manager : [%d cells left]\n",theManager.storageLeft);
    /* return the needed matrix */
    return newMatrix;
  }
  else {
    printf("** Manager : [Out of Memory]\n");
    return NULL;
  }
}


/*********************************************************************/
/* Mtx_Product                                                       */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the product of the given ones.     */
/*********************************************************************/
Mtx_Matrix * Mtx_Product (Mtx_Matrix* mtx1, Mtx_Matrix* mtx2)
{
  int i,j,k;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix(mtx1->nbLine,mtx2->nbColumn);
   
  assert( (*mtx2).nbLine == (*mtx1).nbColumn);

  for(i=0; i<resMtx->nbLine; i++){
    for(j=0; j<resMtx->nbColumn; j++){
      for(k=0; k<mtx1->nbColumn; k++) {
	(*resMtx).tab[i][j] += (*mtx1).tab[i][k] * (*mtx2).tab[k][j];
      }
    }
  }
  return resMtx; 
}

/*********************************************************************/
/* Mtx_Sum                                                           */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the sum of the given ones.         */
/*********************************************************************/
Mtx_Matrix * Mtx_Sum (Mtx_Matrix* mtx1, Mtx_Matrix* mtx2)
{
  int i,j;
  Mtx_Matrix*resMtx  = Mtx_GetNewMatrix (mtx1->nbLine,mtx1->nbColumn);

  assert( (*mtx1).nbLine == (*mtx2).nbLine && (*mtx1).nbColumn == (*mtx2).nbColumn );

  for(i=0; i<mtx1->nbLine; i++){
    for(j=0; j<mtx1->nbColumn; j++){
      (*resMtx).tab[i][j] = (*mtx1).tab[i][j] + (*mtx2).tab[i][j];
    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_Substract                                                     */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the substraction of the given ones */
/*********************************************************************/
Mtx_Matrix * Mtx_Substract (Mtx_Matrix* mtx1, Mtx_Matrix* mtx2)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix(mtx1->nbLine,mtx1->nbColumn);

  assert( (*mtx1).nbLine == (*mtx2).nbLine && (*mtx1).nbColumn == (*mtx2).nbColumn );

  for(i=0; i<mtx1->nbLine; i++){
    for(j=0; j<mtx1->nbColumn; j++){
      (*resMtx).tab[i][j] = (*mtx1).tab[i][j] - (*mtx2).tab[i][j];
    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_CoeffProd                                                     */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the given one multiplied by lambda */
/*********************************************************************/
Mtx_Matrix * Mtx_CoeffProd (Mtx_Matrix* mtx, double lambda)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  for(i=0; i<mtx->nbLine; i++){
    for(j=0; j<mtx->nbColumn; j++){
      (*resMtx).tab[i][j] = lambda * (*mtx).tab[i][j];

    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_AddIdent                                                      */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Add 1 to the diagonal elements                                  */
/*********************************************************************/
Mtx_Matrix * Mtx_AddIdent  (Mtx_Matrix* mtx)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  //assert( (*resMtx).nbLine == (*resMtx).nbColumn );

  for(i=0; i<mtx->nbLine; i++){
    for(j=0; j<mtx->nbColumn; j++){
      if(i==j) (*resMtx).tab[i][j] = 1 +(*mtx).tab[i][j];
      else     (*resMtx).tab[i][j] = (*mtx).tab[i][j];
    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_SubToIdent                                                    */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Subtract the given matrix to the identity matrix.               */
/*********************************************************************/
Mtx_Matrix * Mtx_SubToIdent (Mtx_Matrix* mtx)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  for(i=0;i<mtx->nbLine;i++){
    for(j=0;j<mtx->nbColumn;j++){
      if(i==j) (*resMtx).tab[i][j] = 1 - (*mtx).tab[i][j];
      else     (*resMtx).tab[i][j] =   - (*mtx).tab[i][j];
    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_Neg                                                           */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the negation of the given one.     */
/*********************************************************************/
Mtx_Matrix* Mtx_Neg (Mtx_Matrix* mtx)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  for(i=0; i<mtx->nbLine; i++){
    for(j=0; j<mtx->nbColumn; j++){
      (*resMtx).tab[i][j] = - (*mtx).tab[i][j];

    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_Inverse                                                       */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Assumes that it is a square matrix. Return its inverse.         */
/*********************************************************************/
Mtx_Matrix * Mtx_Inverse (Mtx_Matrix* mtx)
{
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  //assert( (*resMtx).nbLine == (*resMtx).nbColumn );

  if((*resMtx).nbLine==3) {
    double m11 = (*mtx).tab[0][0];
    double m12 = (*mtx).tab[0][1];
    double m13 = (*mtx).tab[0][2];
    double m21 = (*mtx).tab[1][0];
    double m22 = (*mtx).tab[1][1];
    double m23 = (*mtx).tab[1][2];
    double m31 = (*mtx).tab[2][0];
    double m32 = (*mtx).tab[2][1];
    double m33 = (*mtx).tab[2][2];


    double t4  = m11*m22;
    double t6  = m11*m23;
    double t8  = m12*m21;
    double t10 = m13*m21;
    double t12 = m12*m31;
    double t14 = m13*m31;
    double t17 = 1/(t4*m33-t6*m32-t8*m33+t10*m32+t12*m23-t14*m22);

      (*resMtx).tab[0][0] = (m22*m33-m23*m32)*t17;
      (*resMtx).tab[0][1] = (-m12*m33+m13*m32)*t17;
      (*resMtx).tab[0][2] = -(-m12*m23+m13*m22)*t17;
      (*resMtx).tab[1][0] = -(m21*m33-m23*m31)*t17;
      (*resMtx).tab[1][1] = -(-m11*m33+t14)*t17;
      (*resMtx).tab[1][2] = (-t6+t10)*t17;
      (*resMtx).tab[2][0] = (m21*m32-m22*m31)*t17;
      (*resMtx).tab[2][1] = (-m11*m32+t12)*t17;
      (*resMtx).tab[2][2] = -(-t4+t8)*t17;

      return resMtx;
  }

  if((*resMtx).nbLine==2) {
    //  assert((*mtx).tab[0][0]!=0);
    (*resMtx).tab[0][0]=1/(*mtx).tab[0][0];
    return resMtx;
  }

  if((*resMtx).nbLine==1) {
    //  assert((*mtx).tab[0][0]!=0);
    (*resMtx).tab[0][0]=1/(*mtx).tab[0][0];
    return resMtx;
  }  

  return 0;
}

/*********************************************************************/
/* Mtx_Transpose                                                     */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a new matrix equal to the transpose of the given one.    */
/*********************************************************************/
Mtx_Matrix * Mtx_Transpose (Mtx_Matrix* mtx)
{
  int i,j;

  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbColumn,mtx->nbLine);
  
  for(i=0; i<mtx->nbLine; i++){
    for(j=0; j<mtx->nbColumn; j++){
      (*resMtx).tab[j][i] = (*mtx).tab[i][j];
    }
  }
  return resMtx;
}

/*********************************************************************/
/* Mtx_Copy                                                          */
/*                                                                   */
/* SYNOPSIS --                                                       */
/*   Return a copy of the given matrix.                              */
/*********************************************************************/
Mtx_Matrix * Mtx_Copy (Mtx_Matrix* mtx)
{
  int i,j;
  Mtx_Matrix* resMtx = Mtx_GetNewMatrix (mtx->nbLine,mtx->nbColumn);

  for(i=0; i<mtx->nbLine; i++){
    for(j=0; j<mtx->nbColumn; j++){
      (*resMtx).tab[i][j] = (*mtx).tab[i][j];
    }
  }
  return resMtx;
}
