#include <pthread.h>
#include <math.h>
#include "spline.h"
#include "shared.h"

#include "move.h"
#include "i2cManager.h"


#define pi 3.14159265359

#define CHECK_CURRENTMOVE_NNULL

#define MAX_ACCELERATION 0.030

#define CONTROL_CYCLE_PERIOD 100

#define DEBUG_MODULE_NAME "MOVE"
#include "debug.h"

static pthread_t moveThread ;

typedef struct xyMovementStruct
{
  double xTarget;
  double yTarget;
  double lengthEvent;
  pthread_cond_t event;
  pthread_mutex_t mutex;
} xyMovementStruct;

xyMovementStruct xyMovement;

#define STANDBY_MODE 0
#define XY_MODE 1
#define ROTATION_MODE 2
#define LINEAR_MODE 3
#define SPLINE_MODE 4
#define BASIC_LINEAR_MODE 5

int moveMode = STANDBY_MODE;
int updatePosAllowed = 1;
unsigned char emergencyStop = 0;
MovementDescriptor* currentMove = NULL;

int firstPos = 1;
double x=0,y=0,theta=0;
double oldTheta,  oldLengthRight,  oldLengthLeft;

double currentLinearSpeed=0.0;
double currentAngularSpeed=0.0;

int pwmLimitation = 255;
int pwmLimitationActivated = 0;
Event* pwmLimitationEvent;

double getX() { return x; }
double getY() { return y; }
double getTheta() { return theta; }

void setPWMLimitationEvent(Event* event)
{
  pwmLimitationEvent = event; 
}

void setPWMLimitation(int limit)
{
  pwmLimitation = limit;
}

void enablePWMLimitation()
{
  pwmLimitationActivated=1;
}

void disablePWMLimitation()
{
  pwmLimitationActivated=0;
}

// client calls handling

MovementDescriptor* defineBasicLinearMovement(double length, double linearSpeed)
{
  MovementDescriptor* move;
  move=(MovementDescriptor*) malloc(sizeof(MovementDescriptor));
  
  move->moveType=BASIC_LINEAR_MODE;
  move->nbCoords = 4;
  move->coords = (double*) malloc(move->nbCoords * sizeof(double));
  move->coords[0] = x;
  move->coords[1] = y;
  move->coords[2] = length;
  move->coords[3] = linearSpeed;
  
  move->epsilon=3.0;
  move->watchPointList = NULL;
  move->maxSpeedR = 127.0;
  move->maxSpeedL = 127.0;
  move->isMaxProportional=1;
  move->k=0.0;
  move->ki=0.0;
  move->a_ref=0.0;
  move->b_ref=0.0;
  move->endOfMove=NULL;

  return move;
}

MovementDescriptor* defineRotationMovement(double deltaTheta)
{
  int i;
  MovementDescriptor* move;
  
  move=(MovementDescriptor*) malloc(sizeof(MovementDescriptor));

  move->moveType=ROTATION_MODE;
  move->nbCoords = 1;
  move->coords = (double*) malloc(move->nbCoords * sizeof(double));
  move->coords[0] = deltaTheta;
  move->epsilon=0.05;
  move->watchPointList = NULL;
  move->maxSpeedR = 100.0;
  move->maxSpeedL = 100.0;
  move->isMaxProportional=1;
 // move->k=0.2;
  move->k=1.0;
  //move->ki=0.05;
  move->ki=0.001;
  move->a_ref=10.0;
  move->b_ref=0.0;
  move->endOfMove=NULL;
  
  bzero(move->errorMemory, INTEGRAL_MEMORY_SIZE * sizeof(double));
  move->errorMemoryIndex=0;

  return move;
}

void setTargetDist(MovementDescriptor* move, double targetDist)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  if (move->moveType==XY_MODE)
    {
      move->coords[6] = targetDist;
    }
}

void setFinalDist(MovementDescriptor* move, double finalDist)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
    if (move->moveType==XY_MODE)
    {
      move->coords[8] = finalDist;
    }
}

void setFinalK(MovementDescriptor* move, double finalK)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
     if (move->moveType==XY_MODE)
    {
      move->coords[7] = finalK;
    }
}

MovementDescriptor* defineLinearMovement(double length)
{
  MovementDescriptor* move;
  
  move=(MovementDescriptor*) malloc(sizeof(MovementDescriptor));

  move->moveType=LINEAR_MODE;
  move->nbCoords = 4;
  move->coords = (double*) malloc(move->nbCoords * sizeof(double));
  move->coords[0] = length;
  move->coords[1] = 1.0;  // not initialized = 1.0   /  initialiazed = 0.0
  move->coords[2] = 0.0;
  move->coords[3] = 0.0;
  
  move->epsilon=0.05;
  move->watchPointList = NULL;
  move->maxSpeedR = 127.0;
  move->maxSpeedL = 127.0;
  move->isMaxProportional=1;
  move->k=0.1;
  move->a_ref=10.0;
  move->b_ref=0.0;
  move->endOfMove=NULL;
  return move;
}

MovementDescriptor* defineXYMovement(double x, double y)
{
  MovementDescriptor* move;
  
  move=(MovementDescriptor*) malloc(sizeof(MovementDescriptor));

  move->moveType=XY_MODE;
  move->nbCoords = 9;
  move->coords = (double*) malloc(move->nbCoords * sizeof(double));
  move->coords[0] = x;
  move->coords[1] = y;
  move->coords[2] = x;   // start point
  move->coords[3] = y;   // start point
  move->coords[4] = 1.0; // not initialized = 1.0 / initialised = 0
  move->coords[5] = 0.0; // cruise phase = 0.0 / approach phase = 1.0
  move->coords[6]=100.0; // target_dist
  move->coords[7]=0.2; // final_k
  move->coords[8]=60.0; //final_dist
 
  move->epsilon=5.0;
  move->watchPointList = NULL;
  move->maxSpeedR = 127.0;
  move->maxSpeedL = 127.0;
  move->isMaxProportional=1;
  move->k=0.1;
  move->a_ref=10.0;
  move->b_ref=0.0;
  move->endOfMove=NULL;
  return move;
}

MovementDescriptor* defineSplineMovement(Point directionDepart, Point directionFin, Point *points, int nbPoints)
{
  int i;
  MovementDescriptor* move;
  SplineCurve *curve;

  curve = (SplineCurve*) malloc(sizeof(SplineCurve));
  curve->nbPoints = nbPoints;
  curve->directionDepart = directionDepart;
  curve->directionFin = directionFin;
  curve->p = points;

  move = (MovementDescriptor*) malloc(sizeof(MovementDescriptor));

  move->moveType = SPLINE_MODE;
  move->nbCoords = 0; /* on s'en fout plus au moins */
  move->coords = (double*) curve;
  move->maxSpeedL = 127.0;
  move->maxSpeedR = 127.0;
  move->a_ref=10.0;
  move->b_ref=0.0;
  move->k=0.1;
  move->endOfMove=NULL;
  move->watchPointList=NULL;
move->epsilon=5.0;
return move;
}

void setPIDParams(MovementDescriptor* move, double k)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  move->k = k;
}

void setEpsilon(MovementDescriptor* move, double epsilon)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  move->epsilon=epsilon;
}

void setRefPoint(MovementDescriptor* move, double x, double y)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
 move->a_ref=x;
  move->b_ref=y;
}

void setEndMoveEvent(MovementDescriptor* move,  Event* event)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  move->endOfMove = event;
  // lock the event to wait the caller to be ready
  lockEvent(event);
}

void setMaxSpeedInf(MovementDescriptor* move, double maxSpeedR, double maxSpeedL)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  if (move->maxSpeedR > maxSpeedR) move->maxSpeedR = maxSpeedR;
  if (move->maxSpeedL > maxSpeedL) move->maxSpeedL = maxSpeedL;
}

void setMaxSpeed(MovementDescriptor* move, double maxSpeedR, double maxSpeedL, int isProportional)
{
#ifdef CHECK_CURRENTMOVE_NNULL
  if(move==NULL) return;
#endif
  move->maxSpeedR = maxSpeedR;
  move->maxSpeedL = maxSpeedL;
  move->isMaxProportional=isProportional;
}

void startMove(MovementDescriptor* move)
{
  currentMove=move;
  if(currentMove) 
    printf("Switching to mode %d\n",currentMove->moveType);
}

void stopMove()
{
  setSpeed(0.0, 0.0);
  emergencyStop = 1;
}

extern void addWatchPoint(MovementDescriptor* move,
						  Event* event,
					    unsigned char (* testFunction) (double, double, double,void*),
					    void* params)

{
  WatchPoint* watchPoint;

  watchPoint = (WatchPoint*) malloc(sizeof(WatchPoint));
  if(move==NULL || event==NULL) return;
  watchPoint->testFunction=testFunction;
  watchPoint->params=params;

  watchPoint->permanent = 0;

  watchPoint->event = event;

  //  setEventType(watchPoint->event, EVENT_WATCHPOINT);
  //attachEvent(watchPoint->event, (void*)watchPoint);

  watchPoint->next = move->watchPointList;
  move->watchPointList = watchPoint;

  //  return watchPoint;
}

void examineWatchPoints()
{
  WatchPoint* p;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif
  for (p=currentMove->watchPointList ; p != NULL ; p = p->next)
    {
      if  ((p->testFunction)(x,y,theta,p->params))
	{
	  broadcastEvent(p->event);
//	  pthread_cond_broadcast (&(p->event));
	  if (!p->permanent) deleteWatchPoint(currentMove, p);
	}
    }
}

void waitWatchPoint(WatchPoint* watchPoint)
{
  if(watchPoint)
    waitEvent(watchPoint->event);
//  pthread_cond_wait (&(watchPoint->event), &(watchPoint->mutex));
}

void setPermanentWatchPoint(WatchPoint* watchPoint)
{ 
  if(watchPoint)
    watchPoint->permanent = 1;
}

void deleteWatchPoint(MovementDescriptor* move, WatchPoint* watchPoint)
{
}


/*
void defineLinearMovement(double length)
{
  printf("define linear movement...\n");
  xyMovement.xTarget = length;
  xyMovement.yTarget = 0.0;  
}

void goLinearMovement()
{
  moveMode=XY_MODE;
}

void defineLinearMovementEvent(double length)
{
  pthread_cond_init (&xyMovement.event, NULL); 
  pthread_mutex_init (&xyMovement.mutex, NULL);
  xyMovement.lengthEvent = length;
}

void waitLinearMovementEvent()
{
  pthread_cond_wait (&xyMovement.event, &xyMovement.mutex);
}

static void signalLinearMovementEvent()
{
  pthread_cond_broadcast (&xyMovement.event);
}
*/

#define sign(A) (A>=0?1:-1)

void setSpeed(double linearSpeed, double angularSpeed)
{  
  double speedR;
  double speedL;
  double newSpeedR;
  double newSpeedL;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL){
   // printf("SetSpeed, currentmove ==NULL !\n");
     globalState->rightSpeedCons = 0;
     globalState->leftSpeedCons = 0;
     return;
    };
#endif
  currentLinearSpeed = linearSpeed;
  currentAngularSpeed = angularSpeed;

  speedR = (linearSpeed + 2*e*angularSpeed);
  speedL = (linearSpeed - 2*e*angularSpeed);

#if 0
  // max speed

  if (fabs(speedR) > currentMove->maxSpeedR)
    {
      if (currentMove->isMaxProportional)
		speedL=speedL*(currentMove->maxSpeedR/fabs(speedR));      
      speedR=sign(speedR)*currentMove->maxSpeedR;
    }

  if (fabs(speedL) > currentMove->maxSpeedL)
    {
	  if (currentMove->isMaxProportional)
		speedR=speedR*(currentMove->maxSpeedL/fabs(speedL));
      speedL=sign(speedL)*currentMove->maxSpeedL;
    }

#endif

  // max acc

  /*
#ifdef 0
  
  if (speedR > 0)
	{
	  if ((globalState->rightSpeedCons < speedR) && 
		  (((speedR - globalState->rightSpeedCons) / CONTROL_CYCLE_PERIOD) > MAX_ACCELERATION))
		{
		  newSpeedR=(MAX_ACCELERATION*CONTROL_CYCLE_PERIOD) + globalState->rightSpeedCons;
		  if (currentMove->isMaxProportional)
			speedL=speedL*(newSpeedR/speedR);      
		  speedR=newSpeedR;
		}
	}
  
  if (speedL > 0)
	{
	  if ((globalState->leftSpeedCons < speedL) && 
		  (((speedL - globalState->leftSpeedCons) / CONTROL_CYCLE_PERIOD) > MAX_ACCELERATION))
		{
		  newSpeedL=(MAX_ACCELERATION*CONTROL_CYCLE_PERIOD) + globalState->leftSpeedCons;
		  if (currentMove->isMaxProportional)
			speedR=speedR*(newSpeedL/speedL);      
		  speedL=newSpeedL;
		}
	}

#endif
  */

   if (linearSpeed==0.0)
     {
       // rotation
       if (fabs (angularSpeed) > 0)
	 {
	   if ((int)speedR==0)
	     {
	       speedR=2*sign(angularSpeed);
	       speedL=-2*sign(angularSpeed);
	     }
	 }
     }
   
   
   // max speed
   
   if (fabs(speedR) > currentMove->maxSpeedR)
     {
       if (currentMove->isMaxProportional)
	 speedL=speedL*(currentMove->maxSpeedR/fabs(speedR));      
       speedR=sign(speedR)*currentMove->maxSpeedR;
     }
   
   if (fabs(speedL) > currentMove->maxSpeedL)
    {
      if (currentMove->isMaxProportional)
	speedR=speedR*(currentMove->maxSpeedL/fabs(speedL));
      speedL=sign(speedL)*currentMove->maxSpeedL;
    }
   
   
   if (!emergencyStop) {
     globalState->rightSpeedCons = speedR;
     globalState->leftSpeedCons = speedL;
   }
}

int moveToPoint(double xCons, double yCons)
	 /* renvoie 1 si on est arrive au point de consigne (0 sinon) */
{
  double xr;  // relative coords
  double yr;
  
  //  double k11 = 0.1; //0.7
  //double k22 = 0.1; //0.7
  double k11;
  double k22;

  double i11 = 0.0;
  double i22 = 0.0;
  
  double speedR;
  double speedL;

  double sigmaErrorX; //****** TODO ***** define it in the movement structure
  double sigmaErrorY;

  double linearSpeed, angularSpeed;
  double correctedTheta;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return 1;
#endif 
  k11 = currentMove->k;
  k22 = currentMove->k;

  // coordonne de la reference.
  xr = x + currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta);
  yr = y + currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta);

  //on a fini ?
  if ((fabs(xr-xCons)<currentMove->epsilon) 
      && (fabs(yr-yCons)<currentMove->epsilon) )
    
    {
      debug_printf(DEBUG_MODULE_NAME,"Destination reached (%lf - %lf - %lf)...\n",x,y,theta);
      //      printf("Destination reached (%lf - %lf - %lf)...\n",x,y,theta);
      
      globalState->rightSpeedCons = 0;
      globalState->leftSpeedCons = 0;
      return 1;
    }

  linearSpeed=-(i11* sigmaErrorX + k11*(xr-xCons))*cos(theta) - (i22*sigmaErrorX + k22*(yr-yCons))*sin(theta);  

  angularSpeed=(i11* sigmaErrorX + k11*(xr-xCons))/(2*e)*sin(theta)-(i22*sigmaErrorX + k22*((yr-yCons)/(2*e)))*cos(theta);
  
  sigmaErrorX+=xr-xCons;
  sigmaErrorY+=yr-yCons;

  printf("ls : %lf - as : %lf\n",linearSpeed,angularSpeed);

  setSpeed(linearSpeed,angularSpeed);
  return 0;
}

int moveToPoint_ref(double xCons, double yCons)
	 /* renvoie 1 si on est arrive au point de consigne (0 sinon) */
{
  double xr;  // relative coords
  double yr;
  
  //  double k11 = 0.1; //0.7
  //double k22 = 0.1; //0.7
  double k11;
  double k22;

  double i11 = 0.0;
  double i22 = 0.0;
  
  double speedR;
  double speedL;

  double sigmaErrorX; //****** TODO ***** define it in the movement structure
  double sigmaErrorY;

  double linearSpeed, angularSpeed;

  double correctedTheta;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return 1;
#endif
  k11 = currentMove->k;
  k22 = currentMove->k/2;
  // coordonne de la reference.
  //  xr = x;
  //yr = y;
  xr = x + currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta);
  yr = y + currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta);

  //on a fini ?
  if ((fabs(xr-xCons)<currentMove->epsilon) 
      && (fabs(yr-yCons)<currentMove->epsilon) )
    
    {
      debug_printf(DEBUG_MODULE_NAME,"Destination reached (%lf - %lf - %lf - %lf -%lf)...\n",
		   x,y,theta,
		   fabs(xr-xCons),
		   fabs(yr-yCons));
      //      printf("Destination reached (%lf - %lf - %lf)...\n",x,y,theta);
      
      globalState->rightSpeedCons = 0;
      globalState->leftSpeedCons = 0;
      return 1;
    }

  correctedTheta = sign(currentMove->a_ref) * theta;

  linearSpeed=-(k11*(xr-xCons))*cos(correctedTheta) - (k22*(yr-yCons))*sin(correctedTheta);  

  angularSpeed=(k11*(xr-xCons))/(2*e)*sin(correctedTheta)-(k22*((yr-yCons)/(2*e)))*cos(correctedTheta);
  
  //  angularSpeed=sign(currentMove->a_ref) * angularSpeed;

  sigmaErrorX+=xr-xCons;
  sigmaErrorY+=yr-yCons;

  setSpeed(linearSpeed,angularSpeed);
  return 0;
}

double normaliseTheta(double theta)
{
  while (theta<0) { theta += 2*pi; }
  while (theta>=2*pi) { theta -= 2*pi; }
  return theta;
}

double normTheta(double theta, double low)
{
  while (theta<low) { theta += 2*pi; }
  while (theta>=low+2*pi) { theta -= 2*pi; }
  return theta;
}

double tempX=0;
double tempY=0;

void setRobotPos(double newX, double newY, double newTheta)
{
  x = newX;
  y = newY;
  oldTheta = newTheta;
  theta = newTheta;

  tempX = newX;
  tempY = newY;
}



void updatePos()
{  
  int left,right;
  double deltaRight;
  double deltaLeft;
  double newTheta;
  double deltaX, deltaY;

  double tempDX, tempDY, tempR;

  if (firstPos)
    {
      oldLengthRight=K*globalState->rightLength;
      oldLengthLeft=K*globalState->leftLength;
      firstPos=0;
    }
  
  left = globalState->leftLength;
  right = globalState->rightLength; 
  
  deltaRight = K*right - oldLengthRight;
  deltaLeft = K*left - oldLengthLeft;
  
  newTheta=K*(right-left)/(2*e);
  //  newTheta = normaliseTheta(newTheta);

  if (fabs(newTheta-oldTheta) > 0.0001)
    {
      deltaX= 0.5 * (deltaRight + deltaLeft) *
        (sin(newTheta)-sin(oldTheta))/(newTheta-oldTheta);
      deltaY= - 0.5 * (deltaRight + deltaLeft) *
        (cos(newTheta)-cos(oldTheta))/(newTheta-oldTheta);
	  //printf("plus grand "); 

      tempR = (deltaRight + deltaLeft)/(2*(newTheta-oldTheta));
      tempDY = tempR*(1-cos(newTheta-oldTheta))*cos(oldTheta) - tempR * sin(newTheta - oldTheta) * sin (oldTheta);
      tempDX = tempR*(1-cos(newTheta-oldTheta))*sin(oldTheta) + tempR * sin(newTheta - oldTheta) * cos (oldTheta);
    }
  else
  {
      deltaX= 0.5 * (deltaRight + deltaLeft) * 
        cos((newTheta+oldTheta)/2);  
      deltaY= -0.5 * (deltaRight + deltaLeft) * 
        (-sin((newTheta+oldTheta)/2));
	  // printf("plus petit ");

      tempDX=deltaX;
      tempDY=deltaY;
  } 

  tempX += tempDX;
  tempY += tempDY;

  printf("tempX : %lf - tempY : %lf - tempTheta : %lf\n", tempX, tempY,theta);

  // printf ("x, y, deltaX, deltaY, left, right, newTheta, oldTheta, diff: %lf %lf %lf %lf %ld %ld %lf %lf %lf\n", x, y, deltaX, deltaY, left, right, newTheta, oldTheta, newTheta-oldTheta);
  x += deltaX;
  y += deltaY;
  theta = newTheta;
  
  oldTheta = newTheta;
  oldLengthRight = K*right;
  oldLengthLeft = K*left;
}

//#define DIST 100.0
//#define K_FINAL 0.2
//#define DIST_FINAL 60.0

void xyModeControlLoop()
{ 
  double xCons;
  double yCons;
  double dist2;
  double homo_coeff;

  double xn,yn,xr,yr;
  double dist;
  double length;

  double DIST;
  double DIST_FINAL;
  double K_FINAL;

  Event* endOfMove;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif
 DIST = currentMove->coords[6];
  DIST_FINAL = currentMove->coords[8];
  K_FINAL = currentMove->coords[7];
  if (currentMove->coords[4] == 1.0)
    {
      currentMove->coords[4] = 0.0;
      currentMove->coords[2] = x;
      currentMove->coords[3] = y;
    }

  //  xn = x + currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta);
  //yn = y + currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta);

  // estimation des points intermediaires
  dist2 = ((currentMove->coords[0] - x)*(currentMove->coords[0] - x) + 
		   (currentMove->coords[1] - y)*(currentMove->coords[1] - y));

  if ( (dist2 < (DIST_FINAL*DIST_FINAL)) || (currentMove->coords[5]==1.0))
    {
      currentMove->k=K_FINAL;
      xCons = currentMove->coords[0];
      yCons = currentMove->coords[1];
      currentMove->coords[5]=1.0;
      printf("*");
    }
  else
	{
	  //homo_coeff = DIST / sqrt(dist2);
	  //  homo_coeff = DIST / sqrt(dist2);
	  //xCons = xn  + (currentMove->coords[0] - xn) * homo_coeff;
	  //yCons = yn  + (currentMove->coords[1] - yn) * homo_coeff;

	  length = sqrt((currentMove->coords[0] - currentMove->coords[2])*(currentMove->coords[0] - currentMove->coords[2])
			+ (currentMove->coords[1] - currentMove->coords[3])*(currentMove->coords[1] - currentMove->coords[3]));
		    
	  if (length !=0.0)
	    {
	      dist = ( 
		      ( (currentMove->coords[0] - currentMove->coords[2])/length ) * (x - currentMove->coords[2])
		      + 
		      ( (currentMove->coords[1] - currentMove->coords[3])/length ) * (y - currentMove->coords[3])
		      );
	    }
	  else
	    {
	      dist = 0.0;
	    }

	  
	  xCons = currentMove->coords[2] + (((currentMove->coords[0] - currentMove->coords[2])/length) * (dist + DIST));
	  yCons = currentMove->coords[3] + (((currentMove->coords[1] - currentMove->coords[3])/length) * (dist + DIST));
	   
	}

  xr = x + currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta);
  yr = y + currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta);

  printf("x : %lf - y : %lf - theta : %lf |  length : %lf | xc : %lf - yc : %lf  | l : %lf - a : %lf | %d %d\n",
	 x,y,theta, length,
	 xCons,
	 yCons,
	 currentLinearSpeed,
	 currentAngularSpeed,
	 globalState->rightPWM,
	 globalState->leftPWM
	 );

  //  if (moveToPoint(xCons - (currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta)), 
  //				  yCons - (currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta))))
  
  if (moveToPoint_ref(xCons,yCons))
	{
	  debug_printf(DEBUG_MODULE_NAME,"XYMode : destination reached\n");
	  /* destination reached */
	  moveMode=STANDBY_MODE;
	  setSpeed(0.0,0.0);
	  endOfMove=currentMove->endOfMove;
	  currentMove = NULL; // because we have to do this before broadcast
	  if (endOfMove)
		broadcastEvent(endOfMove);
	}
}

void rotationModeControlLoop()
{
  Event* endOfMove;
  
//  double kp = 0.20;
//  double ki = 0.05/INTEGRAL_MEMORY_SIZE;

  double kp;
  double ki ;

  double linearSpeed, angularSpeed;
  double error;
  double sigmaError;

  int i;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif  
  kp = currentMove->k;
  ki  = currentMove->ki;

  setMaxSpeedInf(currentMove,100.0, 100.0);
  printf("%lf - %lf\n",currentMove->maxSpeedR,currentMove->maxSpeedL);

  error=currentMove->coords[0] - normTheta(theta,currentMove->coords[0]-pi);

  // end ?  
  if (fabs(error) < currentMove->epsilon)
    {
      printf("End of rotation movement...\n");
      moveMode=STANDBY_MODE;
      setSpeed(0.0,0.0);
      endOfMove=currentMove->endOfMove;
      currentMove=NULL;
      if (endOfMove)
	    broadcastEvent(endOfMove);
      return;
    }
  
  linearSpeed=0;

  sigmaError = 0.0;
  for (i=0; i < INTEGRAL_MEMORY_SIZE; i++)
    sigmaError+=currentMove->errorMemory[i];

  angularSpeed= kp * error + ki * sigmaError;

  printf("target : %lf - theta : %lf rad - theta : %lf deg- normTheta : %lf - error : %lf - sigmaError : %lf - speed=%lf\n",
	currentMove->coords[0],
	theta,
	 theta * (180/pi),
	 normTheta(theta,currentMove->coords[0]-pi),
	 error,
	 sigmaError,
	 angularSpeed);


  currentMove->errorMemory[(++currentMove->errorMemoryIndex) % INTEGRAL_MEMORY_SIZE] = error;

  setSpeed(linearSpeed, angularSpeed);
}

void basicLinearModeControlLoop()
{
  double xStart ;
  double yStart ;
  double targetLength ;
  double speed;
  double length;
  Event* endOfMove;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif
  xStart = currentMove->coords[0];
  yStart = currentMove->coords[1];
  targetLength = currentMove->coords[2];
  speed = currentMove->coords[3];

  length = sqrt( (xStart - x)*(xStart -x)  +  (yStart - y)*(yStart - y) );
  
  // end ?
  if ( (fabs(  length  - targetLength ) < currentMove->epsilon) || (length > targetLength))
    {
      printf("End of basic linear  movement...\n");
      moveMode=STANDBY_MODE;
      setSpeed(0.0,0.0);
      endOfMove=currentMove->endOfMove;
      currentMove=NULL;
      printf("Start broadcast...\n");
      if (endOfMove)
	    broadcastEvent(endOfMove);
      printf("End broadcast...\n");
      return;
    }

  printf("x : %lf - y : %lf - length : %lf - targetLength : %lf\n",
	 x,
	 y,
	 length,
	 targetLength);

  setSpeed(speed, 0.0);
}


void MoveLinearBasic(double length, int linearSpeed)
{
     double currentLength=0;
     double departX,departY;
     departX=x;
     departY=y;
     while((departX-x)*(departX-x)+(departY-y)*(departY-y)<length*length- 5){ 
         if (!emergencyStop) {
	     globalState->rightSpeedCons = linearSpeed;
	     globalState->leftSpeedCons   = linearSpeed;
	 }
     }
     globalState->rightSpeedCons = 0;
     globalState->leftSpeedCons = 0;
    
}



void linearModeControlLoop()
{
  double targetLength;
  double currentLength;

  double linearSpeed;
  double angularSpeed;

  // transform the linear movement to a xy movement
  //  currentMove->moveType=XY_MODE;
  // free(currentMove->coords);
  // currentMove->nbCoords = 2;
  // currentMove->coords = (double*) malloc(currentMove->nbCoords * sizeof(double));
  // currentMove->coords[0]=x + currentMove->a_ref*cos(theta)-currentMove->b_ref*sin(theta) + length*cos(theta);
  // currentMove->coords[1]=y + currentMove->a_ref*sin(theta)+currentMove->b_ref*cos(theta) + length*sin(theta);
  //printf("Switching to xyMode\n");
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif

  if (currentMove->coords[1] == 1.0) // not initialized
    {
      printf("init\n");
      currentMove->coords[2] = x;
      currentMove->coords[3] = y;
      currentMove->coords[1] = 0.0; // initialized
    }

  targetLength = currentMove->coords[0];
  currentLength = sqrt( ((x - currentMove->coords[2]) * (x - currentMove->coords[2])) 
			+ ((y - currentMove->coords[3]) * (y - currentMove->coords[3])) );
  
  if (currentLength < targetLength - 600)
    linearSpeed = 0.2*(targetLength - 600);
  else
    linearSpeed = 0.2*(targetLength - currentLength);

  angularSpeed = -3.0*theta;

  if (linearSpeed > currentLinearSpeed + 20)  linearSpeed=currentLinearSpeed + 20;

  printf("x : %lf - y : %lf - theta %lf -- tl : %lf -- cl : %lf -- ls : %lf -- rs : %lf\n",
	  x,
	  y,
	  theta,
	  targetLength,
	  currentLength,
	  linearSpeed,
	  angularSpeed);

  setSpeed(linearSpeed, angularSpeed);
}

static double t = 0; /* point actuel sur la courbe */
void splineModeControlLoop()
{
  SplineCurve *curvePtr;
  SplineCurve curve;
  int maxT;
  int arrive = 0;
  Event* endOfMove;
  Point p;
  p.x =x;
  p.y =y;
#ifdef CHECK_CURRENTMOVE_NNULL
  if(currentMove==NULL) return;
#endif

  curvePtr = (SplineCurve*) (currentMove->coords);
  curve = *curvePtr;
  maxT = curve.nbPoints - 1;

  t = getNextPoint(curve, p, t);

  spline(curve, t, &p);
  arrive = moveToPoint(p.x, p.y);

  printf("x : %lf - y : %lf - theta : %lf - xc : %lf - yc : %lf\n",
	 x,y,theta,
	 p.x,p.y);


  if ((t>=maxT) && (arrive))
	{
	  /* destination reached */
      printf("End of movement...\n");
//	  free(curve.p);
	  t = 0; /* pour le prochain mouvement */
	  
	  moveMode=STANDBY_MODE;
	  setSpeed(0.0,0.0);
	  endOfMove=currentMove->endOfMove;
	  currentMove=NULL;
	  if (endOfMove)
	    broadcastEvent(endOfMove);
      return;
	}
}

void controlLoop()
{
  if (emergencyStop)
    {
      setSpeed(0.0,0.0);
      currentMove=NULL;
      emergencyStop = 0;
    }
  if (currentMove)
    {
      printf("currentMove\n");
      switch (currentMove->moveType)
		{
		case ROTATION_MODE : 
		  rotationModeControlLoop();
		  break;     
		  
		case LINEAR_MODE : 
		  linearModeControlLoop();
		  break;     

		case BASIC_LINEAR_MODE :
		  basicLinearModeControlLoop();
		  break;
		
		case XY_MODE : 
		  xyModeControlLoop();
		  break;

		case SPLINE_MODE : 
		  //printf(".\n");
		  splineModeControlLoop();
		  break;

		default :
		  printf("x : %lf - y : %lf - theta : %lf\n",x,y,theta); 

		}
    }
  else
    {
      //printf("x : %lf - y : %lf - theta : %lf\n",x,y,theta); 
    }
}

void resetStop()
{
  updatePosAllowed=0; // disable auto update of the position

  updatePos();

  INIThctl();

//  i2cSend(I2C_INITHCTL_SLAVE, 0, 1 << I2C_INITHCTL_BIT);
//  stopMove();
//  usleep(500000);
//  i2cSend(I2C_INITHCTL_SLAVE, 1<< I2C_INITHCTL_BIT, 1 << I2C_INITHCTL_BIT);
//  waitHCTLSoftInit();

  firstPos=1; // to reset the old pos in position computation
  updatePos();

  updatePosAllowed=1; // enable autoupdate
}

void *moveThreadBody()
{
  printf("In thread body\n");
  while (1)
    {
      //printf("update pos\n");
      if (updatePosAllowed)
	{
	  updatePos();
	}
      //printf("controlLoop\n");
	  controlLoop();
	  
//	  printf("%lf - %lf - %lf\n",x,y,theta);
 //printf("examineWatchPoints\n");
	  if (currentMove)
	    examineWatchPoints();
  //printf("pwmLimitationActivated\n");
	  if (pwmLimitationActivated)
	    {
	      if ( (abs(globalState->rightPWM) > pwmLimitation) || (abs(globalState->leftPWM) > pwmLimitation) )
		{
		  broadcastEvent(pwmLimitationEvent);
		}
	    }

      usleep(CONTROL_CYCLE_PERIOD);
    }
}

int move_initModule()
{
  printf("Init move module\n");
  
  // Create the move thread
  return pthread_create(&moveThread, NULL, moveThreadBody,NULL);
}
