#include "events.h"

#define DEBUG_MODULE_NAME EVENTS
#include "debug.h"

typedef struct twoEvents
{
  Event* event1;
  Event* event2;
} twoEvents;

typedef struct threeEvents
{
  Event* event1;
  Event* event2;
  Event* event3;  
} threeEvents;

void createEvent(Event* event, char* eventName)
{
  pthread_cond_init (&(event->cond), NULL);
  pthread_mutex_init (&(event->condMutex),NULL);

  event->event1   = NULL;
  event->event2   = NULL;
  event->joinType = EVENT_JOIN_NO;
  strcpy(event->eventName,eventName);
}

void broadcastEvent(Event* event)
{
  pthread_mutex_lock(&(event->condMutex));
  pthread_cond_broadcast (&(event->cond));
  pthread_mutex_unlock(&(event->condMutex));
}

void* waitingOrThreadBodyCancelProc(twoEvents* events)
{
  debug_printf(DEBUG_MODULE_NAME,"Cancel OR thread (0x%x)\n",pthread_self());
  pthread_mutex_unlock(&(events->event1->condMutex));
  return NULL;
}

void* waitingOrThreadBody(twoEvents* events)
{
  int oldCancelType;
  
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,&oldCancelType);

  pthread_cleanup_push (waitingOrThreadBodyCancelProc, events);
  
  debug_printf(DEBUG_MODULE_NAME,"In thread OR body(ev1 : %s - ev2 : %s - 0x%x)\n",
			   events->event1->eventName,
			   events->event2->eventName,
			   pthread_self());
  
  waitEvent(events->event1);

  debug_printf(DEBUG_MODULE_NAME,"Joined event (%s) arrived (thread 0x%x) !!!!\n",
			   events->event1->eventName,
			   pthread_self());

  events->event2->signalingEvent = events->event1;

  broadcastEvent(events->event2);

  pthread_cleanup_pop (0);

  return NULL;
}

void* waitingAndThreadBodyCancelProc(twoEvents* events)
{
  debug_printf(DEBUG_MODULE_NAME,"Cancel AND thread (0x%x)\n",pthread_self());
  pthread_mutex_unlock(&(events->event1->condMutex));
  pthread_mutex_unlock(&(events->event2->condMutex));
  return NULL;
}

void* waitingAndThreadBody(threeEvents* events)
{
  int oldCancelType;
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,&oldCancelType);

  pthread_cleanup_push (waitingOrThreadBodyCancelProc, events);

  debug_printf(DEBUG_MODULE_NAME,"In thread AND body(0x%x - 0x%x - 0x%x - 0x%x)\n",
			   events->event1,
			   events->event2,
			   events->event3,
			   pthread_self());

  waitEvent(events->event1);

  debug_printf(DEBUG_MODULE_NAME,"Joined event 1 arrived (thread 0x%x) !!!!\n",
			   pthread_self());

  waitEvent(events->event2);

  debug_printf(DEBUG_MODULE_NAME,"Joined event 2 arrived (thread 0x%x) !!!!\n",
			   pthread_self());  

  broadcastEvent(events->event3);

  pthread_cleanup_pop (0);

  return NULL;
}

void waitEventNoJoin(Event* event)
{
  debug_printf(DEBUG_MODULE_NAME,"Waiting condition (%s)\n",event->eventName);
  pthread_mutex_lock(&(event->condMutex));
  debug_printf(DEBUG_MODULE_NAME,"Waiting (has mutex) condition (%s)\n",event->eventName);
  pthread_cond_wait (&(event->cond), &(event->condMutex));
  pthread_mutex_unlock(&(event->condMutex));
  debug_printf(DEBUG_MODULE_NAME,"Condition arrived (%s)\n",event->eventName);
}

void waitEventOrJoin(Event* event)
{
  pthread_t thread1;
  pthread_t thread2;
  twoEvents* events;


  debug_printf(DEBUG_MODULE_NAME,"Creating thread 1\n");
  events = (twoEvents*)malloc(sizeof(twoEvents));
  events->event1 = event->event1;
  events->event2 = event;
  pthread_create(&thread1, NULL, (void *(*)(void *))waitingOrThreadBody, events);
  
  debug_printf(DEBUG_MODULE_NAME,"Creating thread 2\n");
  events = (twoEvents*)malloc(sizeof(twoEvents));
  events->event1 = event->event2;
  events->event2 = event;
  pthread_create(&thread2, NULL, (void *(*)(void *))waitingOrThreadBody, events);
  
  debug_printf(DEBUG_MODULE_NAME,"Waiting on %s\n",event->eventName);
  pthread_mutex_lock(&(event->condMutex));
  pthread_cond_wait (&(event->cond), &(event->condMutex));
  pthread_mutex_unlock(&(event->condMutex));
  debug_printf(DEBUG_MODULE_NAME,"Event %s arrived\n",event->eventName);

  pthread_cancel(thread1);
  pthread_cancel(thread2);


  /*  pthread_cond_destroy(&event->cond);
  pthread_mutex_destroy(&event->cond);
  pthread_cond_init (&(event->cond), NULL);
  pthread_mutex_init (&(event->condMutex),NULL);

  pthread_cond_destroy(&(event->event1->cond));
  pthread_mutex_destroy(&(event->event1->condMutex));
  pthread_cond_init (&(event->event1->cond), NULL);
  pthread_mutex_init (&(event->event1->condMutex),NULL);

  pthread_cond_destroy(&(event->event2->cond));
  pthread_mutex_destroy(&(event->event2->condMutex));
  pthread_cond_init (&(event->event2->cond), NULL);
  pthread_mutex_init (&(event->event2->condMutex),NULL);*/
  
}


void waitEventAndJoin(Event* event)
{  
  pthread_t thread1;
  pthread_t thread2;
  threeEvents* events;

  debug_printf(DEBUG_MODULE_NAME,"Creating thread 1\n");
  events = (twoEvents*)malloc(sizeof(twoEvents));
  events->event1 = event->event1;
  events->event2 = event->event2;
  events->event3 = event;
  pthread_create(&thread1, NULL, (void *(*)(void *))waitingAndThreadBody, events);
  
  debug_printf(DEBUG_MODULE_NAME,"Creating thread 2\n");
  events = (twoEvents*)malloc(sizeof(twoEvents));
  events->event1 = event->event2;
  events->event2 = event->event1;
  events->event3 = event;
  pthread_create(&thread2, NULL, (void *(*)(void *))waitingAndThreadBody, events);
  
  pthread_mutex_lock(&(event->condMutex));
  pthread_cond_wait (&(event->cond), &(event->condMutex));
  pthread_mutex_unlock(&(event->condMutex));
  
  pthread_cancel(thread1);
  pthread_cancel(thread2);
}

void waitEvent(Event* event)
{
  switch (event->joinType)
	{
	case EVENT_JOIN_NO :
	  waitEventNoJoin(event);
	  break;
	case EVENT_JOIN_OR :
	  waitEventOrJoin(event);
	  break;
	case EVENT_JOIN_AND :
	  waitEventAndJoin(event);
	  break;	  
	}
}

void orEvents(Event* event1, Event* event2, Event* eventResult)
{
  eventResult->event1=event1;
  eventResult->event2=event2;
  eventResult->joinType = EVENT_JOIN_OR;
}

void andEvents(Event* event1, Event* event2, Event* eventResult)
{
  eventResult->event1=event1;
  eventResult->event2=event2;
  eventResult->joinType = EVENT_JOIN_AND;
}
