#include <stdio.h>

#include "linux/i2c.h"
#include "linux/i2c-dev.h"

#include "i2cManager.h"

#define DEBUG_MODULE_NAME I2C
#include "debug.h"



#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <pthread.h>
#include <fcntl.h>
#include <stdlib.h>

static int i2c_descriptor;

static pthread_t i2cThread ;
static pthread_cond_t i2cEvent;
static pthread_mutex_t i2cEventMutex;

static pthread_mutex_t cmdListMutex;

static unsigned char i2cSlaveState[NB_I2C_SLAVES];

#define I2C_NB_TRY_MAX 10

#define TYPE_COMMAND 0
#define TYPE_ACQUISITION 1



typedef struct i2cCommand
{
  unsigned char cmdType;
  unsigned char slave;
  unsigned char slave_id;
  unsigned char result;
  unsigned char content; // the cmd send or the response received
  unsigned char mask;
  unsigned char ttl;
  pthread_cond_t endEvent;
  pthread_mutex_t endEventMutex;
  struct i2cCommand* next;
} i2cCommand;

i2cCommand* cmdList = NULL;

void addCommandStruct(i2cCommand* cmd)
{
  i2cCommand* indexCmd;

  pthread_mutex_lock(&cmdListMutex);

  cmd->next=NULL;

    if (cmdList != NULL)
    {
      indexCmd = cmdList;
      while (indexCmd->next != NULL) indexCmd = indexCmd->next;      
      indexCmd->next=cmd;
    }  
  else
    {
      cmdList = cmd;
    }

  pthread_mutex_unlock(&cmdListMutex);
    
  pthread_cond_broadcast (&i2cEvent);
}

i2cCommand* addCommand(unsigned char type, unsigned char slave, unsigned char value, unsigned char mask, unsigned char ttl)
{
  i2cCommand* cmd;
  i2cCommand* indexCmd;
  int i;

  cmd = (i2cCommand*)malloc(sizeof(i2cCommand));
  cmd->cmdType=type;
  cmd->slave=slave;
  cmd->content=value;
  cmd->mask=mask;
  cmd->next=NULL;
  cmd->ttl=ttl;

  for (i=0 ; i<NB_I2C_SLAVES; i++)
    if (I2C_SLAVE_ADDRESS[i]==slave) cmd->slave_id=i;
  
  
  if (cmd->cmdType == TYPE_ACQUISITION)
    {
      pthread_cond_init (&(cmd->endEvent), NULL); 
      pthread_mutex_init (&(cmd->endEventMutex),NULL);

	  // !!!!!!!!!!!!!!!!!! GET MUTEX !!!!!!!!!!!!!!!!
      pthread_mutex_lock(&(cmd->endEventMutex));
    }

  addCommandStruct(cmd);

  return cmd;
}

int i2cGet(unsigned char slave, unsigned char mask)
{
  i2cCommand* cmd;
  int result;

  debug_printf(DEBUG_MODULE_NAME,"Filling fifo...\n"); 

  cmd = addCommand(TYPE_ACQUISITION, slave, 0, mask,  I2C_NB_TRY_MAX);

  
  debug_printf(DEBUG_MODULE_NAME,"Waiting for acquisition completion...\n"); 

  // mutex already locked by addCommand
  pthread_cond_wait (&(cmd->endEvent), 
		     &(cmd->endEventMutex));

  // BUG ? //   pthread_mutex_unlock(&(cmd->endEventMutex));

  if (cmd->result==0)
    result = cmd->content & cmd->mask;
  else
    result = cmd->result;

  pthread_cond_destroy(&(cmd->endEvent));
  pthread_mutex_destroy(&(cmd->endEventMutex));

  debug_printf(DEBUG_MODULE_NAME,"Return acquisition result on slave 0x%x (slave value : %d - value : %d)\n",
	       cmd->slave,
	       cmd->content,
	       result);

  free(cmd);

  return result;
}

void i2cReAcq(unsigned char slave, unsigned char value, unsigned char mask, unsigned char ttl)
{
  addCommand(TYPE_ACQUISITION, slave, value, mask,  ttl);
}

void i2cReSend(unsigned char slave, unsigned char value, unsigned char mask, unsigned char ttl)
{
  addCommand(TYPE_COMMAND, slave, value, mask,  ttl);
}

void i2cSend(unsigned char slave, unsigned char value, unsigned char mask)
{
  i2cReSend(slave,value,mask,I2C_NB_TRY_MAX);
}

void handleCommandSend(i2cCommand *cmd)
{
  int ok = 1;
  char dummy;
  unsigned char command;

  debug_printf(DEBUG_MODULE_NAME,"Processing command on slave 0x%x with (value 0x%x - mask 0x%x)\n",
	       cmd->slave,
	       cmd->content,
	       cmd->mask);

  if (ioctl(i2c_descriptor, I2C_SLAVE, cmd->slave) < 0) ok=0;    

  i2cSlaveState[cmd->slave_id] = (i2cSlaveState[cmd->slave_id] & ~cmd->mask) | cmd->content;
 
  if (write(i2c_descriptor,&i2cSlaveState[cmd->slave_id], 1) != 1)  ok=0;

  if (!ok)
  {
    if (cmd->ttl>0)
      {
	debug_printf(DEBUG_MODULE_NAME,"Unable to send the command (%s), retry it later (try %d of %d)\n", 
		     strerror(errno),
		     I2C_NB_TRY_MAX-cmd->ttl,
		     I2C_NB_TRY_MAX);
	
	i2cReSend(cmd->slave, cmd->content, cmd->mask,--cmd->ttl);
      }
    else
      {
	debug_printf(DEBUG_MODULE_NAME,"Unable to send the command (%s) after %d tries abondon !!!\n", 
		     strerror(errno),
		     I2C_NB_TRY_MAX);	
      }
  }

}

void handleCommandRecv(i2cCommand *cmd)
{
  int ok;

    debug_printf(DEBUG_MODULE_NAME,"Processing acquisition on slave 0x%x with (mask 0x%x)\n",
	       cmd->slave,
	       cmd->mask);

  if (ioctl(i2c_descriptor, I2C_SLAVE, cmd->slave) < 0) ok=0;    

  if (read(i2c_descriptor,&(cmd->content), 1) != 1)  ok=0;

  if (!ok)
    {
      if (cmd->ttl>0)
		{
		  debug_printf(DEBUG_MODULE_NAME,"Unable to acquire (%s), retry it later (try %d of %d)\n", 
					   strerror(errno),
					   I2C_NB_TRY_MAX-cmd->ttl,
					   I2C_NB_TRY_MAX);
		  
		  //		  cmd->ttl--;
		  i2cReAcq(cmd->slave, cmd->content, cmd->mask,--cmd->ttl);
		  //		  addCommandStruct(cmd);
		}
      else
		{
		  debug_printf(DEBUG_MODULE_NAME,"Unable to acquire the command (%s) after %d tries abondon !!!\n", 
					   strerror(errno),
					   I2C_NB_TRY_MAX);	
		  cmd->result=-1;

  		  // BUG ? //pthread_mutex_lock(&(cmd->endEventMutex));
		  
		  // BUG ? //pthread_mutex_unlock(&(cmd->endEventMutex));
		  pthread_cond_broadcast (&(cmd->endEvent));
		}
    }
  else
    {
      // update the state of the slave
      i2cSlaveState[cmd->slave_id]=cmd->content;
	  
      cmd->result=0;


      ///////////////////////// ******************** BUG ???????????????????
      //pthread_mutex_lock(&(cmd->endEventMutex));

      //  pthread_mutex_unlock(&(cmd->endEventMutex));

      debug_printf(DEBUG_MODULE_NAME,"Signaling end of acquisition\n");

      pthread_mutex_lock(&(cmd->endEventMutex));
      pthread_cond_broadcast (&(cmd->endEvent));
      pthread_mutex_unlock(&(cmd->endEventMutex));
    }
}

void handleCommand(i2cCommand *cmd)
{
  if (cmd->cmdType==TYPE_COMMAND) handleCommandSend(cmd);
  if (cmd->cmdType==TYPE_ACQUISITION) handleCommandRecv(cmd);
}

void *i2cThreadBody()
{
  i2cCommand *cmd;

  sleep(1);
  
  while(1)
    {
      debug_printf(DEBUG_MODULE_NAME,"Waiting for command...\n");
	  pthread_mutex_lock(&i2cEventMutex);
      pthread_cond_wait (&i2cEvent, &i2cEventMutex);
      debug_printf(DEBUG_MODULE_NAME,"Cmd received...\n");
	  pthread_mutex_unlock(&i2cEventMutex);
	  
      do {
	pthread_mutex_lock(&cmdListMutex);
	cmd = cmdList;
	if (cmd)
	  cmdList = cmdList->next; 
	
	pthread_mutex_unlock(&cmdListMutex);
	
	if (cmdList==NULL)
	  debug_printf(DEBUG_MODULE_NAME,"Cmd FIFO empty...\n");

	if (cmd)
	  handleCommand(cmd);
	
	//	if (cmd->cmdType==TYPE_COMMAND)
	//  {
	//    free(cmd);
	//  }

      } while (cmdList !=NULL);
    }
}



int i2c_initModule()
{
  int i;
  int result = 0;

  debug_printf(DEBUG_MODULE_NAME,"Init module\n");

  if ((i2c_descriptor = open("/dev/i2c-0",O_RDWR)) < 0)
    {
      debug_printf(DEBUG_MODULE_NAME,"Unable to open /dev/i2c-0 (%s)\n",strerror(errno));
      result = -1;
    }

  ioctl(i2c_descriptor,I2C_TENBIT,0);

  if (!pthread_create(&i2cThread, NULL, i2cThreadBody,NULL))
    {
      debug_printf(DEBUG_MODULE_NAME,"Unable to create the i2c thread (%s)\n",strerror(errno));
      result = -1;
    }

  pthread_cond_init (&i2cEvent, NULL); 
  pthread_mutex_init (&i2cEventMutex, NULL);

  pthread_mutex_init (&cmdListMutex,NULL);

  sleep(2);

  for (i=0 ; i<NB_I2C_SLAVES ; i++)
    {
      i2cSlaveState[i] = i2cGet(I2C_SLAVE_ADDRESS[i],0xFF);
	  i2cSend(I2C_SLAVE_ADDRESS[i],0xFF,i2cSlaveState[i]);
      debug_printf(DEBUG_MODULE_NAME,"Getting initial state from slave 0x%x -> 0x%x\n", 
		   I2C_SLAVE_ADDRESS[i],
		   i2cSlaveState[i]);
    }

  return result;
}

void i2c_cleanUpModule()
{
  debug_printf(DEBUG_MODULE_NAME,"cleanUp module\n");
}









